package universe

import plant._
import animal._

class World(val width:Int=20,val height:Int=20) {
  val board = new Board(width,height)
  val flora = new scala.events.VarList[Plant] 
  val animals = new scala.events.VarList[Animal] 

  evt plantDied = flora.any(p => p.disappeared)  
  evt plantGerminated = flora.any(p => p.germinated)  

  evt animalDied = animals.any(p => p.disappeared)  
  evt animalBorn = animals.any(p => p.born)  

  var hour : Int =0
  var day : Int = 0
  var week : Int =0

  imperative evt clock[Unit]
  evt newDay[Unit] = clock && (() => hour == 0)
  evt newWeek[Unit] = newDay && (() => day == 0)
  evt newMonth[Unit] = newWeek && (() => week == 0)

  board.addedElement += elementAdded _
  clock += nextHour _
  newDay += nextDay _
  newWeek += nextWeek _
  newMonth += nextMonth _ 
  animalDied += onAnimalDied _
  plantDied += onPlantDied _

  def elementAdded(be: BoardElement) = be match {
    case a: Animal =>
      animals += a
    case p: Plant =>
      flora += p
  }

  def onAnimalDied(a: Animal) {
    animals -= a
  }

  def onPlantDied(p: Plant) {
    flora -= p
  }

  def dump = {
    println("Dump "+board.surface.size + ". Week: " + week + ", Day: " + day + ", Hour: " + hour)
    for (y <- 0 to height) {
      var line=""
      var img=""
      for (x <- 0 to width) {
        val e=board(x,y)
        if (e!=null) {
          line+=" "+e.toString
          img+=e.symbol
        } else {
          img += " "
        }
      }
//      println("|"+img+"|"+line)
      println("|"+img+"|")
    }
  }

  def nextHour() {
    hour = (hour + 1) % 24
  }

  def nextDay() {
    day = (day + 1) % 7
  }

  def nextWeek() {
    week = (week + 1) % 5
    new Plant(this, null, 150)
  }

  def nextMonth() {
  }

}

object Test {
  val wrld=new World(50,25)
  val gardener=new Gardener(wrld)
  def putH = {
    val board=wrld.board
    val center=new Pos(board.width/2,board.height/2)
    val inner=if (board.width>board.height) board.height else board.width
    for (i <-1 to 5) {
      val m_pos=center.random_around(3,inner/3)
      val f_pos=center.random_around(3,inner/3)
      
      board.moveTo(m_pos,
		   new Herbivore(wrld,5,70) with HerbivoreMale)
      board.moveTo(f_pos,
		   new Herbivore(wrld,3,80) with HerbivoreFemale)
      board.moveTo(center.random_around(inner/2, inner - 1),
        new Carnivore(wrld, 5, 100) with CarnivoreMale)
      board.moveTo(center.random_around(inner/2, inner -1),
        new Carnivore(wrld, 5, 100) with CarnivoreFemale)
      new Plant(wrld,null,80)
      new Plant(wrld,null,60)
      new Plant(wrld,null,40)
    }
  }
  def main(a : Array[String]) = {
    putH
    while (true) {
      wrld.clock()
      wrld.dump
      Thread.sleep(50)
    }
  }
}
