package universe
package animal

import scala.util._

object Carnivore {
  def asCarnivore(be:BoardElement) : Option[Carnivore] = be match {
    case c: Carnivore => Some(c)
    case _ => None
  }
}

abstract class Carnivore(world: World,range:Int=10, maxAge:Int=160) extends Animal(world,range,maxAge) with Reproducible {

  evt herbivoreReached[Animal] = otherSpeciesReached && (_.isInstanceOf[Herbivore])

  override evt movementTick[Unit] = super.movementTick && (() => !needSleep)

  def sameSpecies(a: Animal) = a.isInstanceOf[Carnivore]

  // when hungry the prey doesn't need to be big at all
  def check_hungry(a : Animal) : Unit = a match {
    case h: Herbivore if target() == null && energy < 40 && distance(h) < range =>
      target := h
    case _ => // do nothing
  }

  def find_food(around : List[BoardElement]):BoardElement = {
    val prey = around.filter(_ match {
      case x: Herbivore =>
        x.energy < energy
      case _ => false
    }).map(be => Herbivore.asHerbivore(be).get)

    if (prey.size > 0) {
      val pos=scala.math.sqrt(Random.nextInt(prey.size*prey.size)).toInt
      // we prefer a big prey at close distance
      val value_ordered_prey=prey.sortWith(
        (p1,p2) => p1.energy/(1+distance(p1)) < p2.energy/(1+distance(p2))
      )
      value_ordered_prey(pos)
    } else {
      null
    }
  }

  var sleep=1

  // kind of pendulum to controle sleep or awake
  def needSleep() = {
    // sleep should never become 0
    val sign=sleep/sleep.abs
    if (sleep.abs==1) {
      if (sleep==1) {
        sleep-=2+Random.nextInt(6)
      } else {
        sleep+=8+Random.nextInt(15)
      }

    } else {
      sleep -= sign
    }

    sleep >= 0
    false
 }


  override observable def appear() = {
    super.appear
    world.animalBorn += check_hungry _
    herbivoreReached += eatHerbivore _
    energy = 50
  }

  override observable def die() {
    world.animalBorn -= check_hungry _
    herbivoreReached -= eatHerbivore _
    super.die
  }
  
  override def find_target(range :Int):BoardElement  = {
    val around = world.board.neighbours(this,range)
    val like_food=if (energy<100) 1 else 0
    val like_partner=if (this.isFertile) 1 else 0   
    val need_partner=if (age>90 && this.isFertile) 1 else 0   
    
    // specify the probability for what the herbivore
    // will be looking for
    val food_prob=5+like_food*35
    val partner_prob=1+like_partner*30+need_partner*80
    val tot_prob=food_prob+partner_prob
    // repeat a few times to make it more probably to find a target
    for (i<-1 to 3) {
      val r=Random.nextInt(tot_prob+1)
      var b : BoardElement =null
      if (b==null && r < food_prob)
        b=find_food(around)
      if (b==null && r < food_prob+partner_prob)
        b=find_partner(around)
      if (b!=null)
        return b
    }
    // ultimate attempt to find food ( in the unlikely case all 
    // attempts before tried to find a partner
    return find_food(around)
  }


  def eatHerbivore(target: Animal) {
    energy += 2 * target.energy
    target.killed()
  }   

}



