package universe
package animal

import scala.util._

trait Reproducible extends InteractWithAnimals {

  def haveSexWith(a: Animal) {
    val pair: Option[(Male, Female)] = (this, a) match {
      case (f: Female, m: Male) => Some((m, f))
      case (m: Male, f: Female) => Some((m, f))
      case _ => None
    }
    pair match {
      case Some((m,f)) if f.isFertile && m.isFertile => f.conceiveChildWith(m)
      case _ => // do nothing
    }
  }

  override observable def appear() {
    super.appear()
    sameSpeciesReached += haveSexWith _
  }

  override observable def die() {
    sameSpeciesReached -= haveSexWith _
    super.die()
  }


  // a partner must be from the same species
  // it must be fertile and of the opposite sex
  def find_partner(around : List[BoardElement]):BoardElement = {
    val partners = around.filter {
      case a: Animal => sameSpecies(a) && a != this && a.isMale != isMale && a.isFertile
      case _ => false
    } map (_.asInstanceOf[Animal])

    if (partners.size>0) {
      val pos=scala.math.sqrt(Random.nextInt(partners.size*partners.size)).toInt
      val value_ordered_partners=partners.sortWith(
        (p1,p2) => p1.age * p1.energy / (1 + distance(p1)) < p2.age * p2.energy / (1 + distance(p2))
      )
      value_ordered_partners(pos)
    } else {
      null
    }
  }
}

// vim: set ts=4 sw=4 et:
