package universe
package animal

import scala.util._
import plant._

object Herbivore {

  def asHerbivore(be: BoardElement): Option[Herbivore] = be match {
    case h: Herbivore => Some(h)
    case _ => None
  }

}

abstract class Herbivore(world: World,range:Int=10, maxAge:Int=160) extends Animal(world,range,maxAge) with Reproducible {

  evt plantReached[Plant] = (targetReached && (_.isInstanceOf[Plant])) map ((be: BoardElement) => Plant.asPlant(be).get)

  override observable def appear() {
    super.appear
    world.plantGerminated += check_hungry _
    plantReached += onPlantReached _
  }

  def sameSpecies(a: Animal) = a.isInstanceOf[Herbivore]

  def check_hungry(p : Plant) {
    if (energy < 10 && target() == null && distance(p) < range) {
      target := p
    }
  }

  def find_food(around: List[BoardElement]):BoardElement = {
    val minSize=5+(30/(1+energy.abs))
    val food=around.filter(_ match {
      case p: Plant => p.size > minSize
      case _ => false
    })

    if (food.size>0) {
      val g=scala.math.sqrt(Random.nextInt(food.size*food.size)).toInt
      food(g)
    } else {
      null
    }
  }

  override observable def die() {
    world.plantGerminated -= check_hungry _
    plantReached -= onPlantReached _
    super.die
  }

  override def find_target(range :Int):BoardElement  = {
    val around=world.board.neighbours(this,range)
    val like_food=if (energy<30) 1 else 0
    val like_partner=if (this.isFertile) 1 else 0   
    val need_partner=if (age>70 && this.isFertile) 1 else 0   
    
    val food_prob=1+like_food*20
    val partner_prob=1+like_partner*30+need_partner*40
    val tot_prob=food_prob+partner_prob

    for (i<-1 to 3) {
      val r=Random.nextInt(tot_prob+1)
      var b : BoardElement =null
      val around=world.board.neighbours(this,range)
      
      if (b==null && r < food_prob)
        b=find_food(around)
      
      if (b==null && r < food_prob+partner_prob)
        b=find_partner(around)

      if (b!=null)
        return b
    }
  
    return null
  }

  def onPlantReached(p: Plant) {
    val food_desire:Int=60 / (1 + energy.abs) + age * 4
    energy += p.drain(10 + Random.nextInt(1 + food_desire))
  }


  override def toString =
    symbol + "(" + (energy.toFloat/10) + ")#" + age + "@" + pos

}
