package universe
package plant

object Plant {

  def asPlant(b:BoardElement): Option[Plant] = b match {
    case p: Plant => Some(p)
    case _ => None
  }
 
}

class Plant(val world: World,val near:Plant =null,var maxAge:Int=160) extends BoardElement {
  val plnt_rnd = new scala.util.Random

  init

  evt germinated[Plant] = afterExec(plant) && ((_: Pos, success: Boolean) => success) map ((_: Any) => this)
  
  evt disappeared[Plant] = afterExec(die) map ((_: (Unit, Unit)) => this)

  /** Internal event to unregister everything that need to be unregistered before notifying the death to outer */
  protected evt shouldDie[Unit] = (afterExec(get_older) || afterExec(grow)) && (_ => !canLive) || killed

  imperative evt killed[Unit]
  
  protected evt shouldExpand[Unit] = afterExec(get_older) && ((_: Any) => (age > 10 && size > 60 && plnt_rnd.nextInt(50) >= 20))

  var age = 0
  var size = 1 
  override def move(r: Int) = false

  def appear() {
    world.clock += grow _
    world.newDay += get_older _
    shouldExpand += expand _
    shouldDie += die _
  }

  def drain(amount : Int) = {
    if (amount>size) {
      val oldSize=size
      size=0
      canLive
      oldSize
    } else {
      size -= amount
      amount
    }
  }

  def init {
    var attempts = 10
    while (attempts > 0) {
      attempts -= 1
      if (near != null) {
        val freepos = near.find_free_around(4 + size / 5, 2 + size / 8)
        if (freepos == null) {
          near.age += 50
          attempts = 0
        } else if (plant(freepos)) {
          attempts = 0
        } 
      } 
      if (near == null) {
	      val freepos = new Pos(plnt_rnd.nextInt(world.board.width),
			                        plnt_rnd.nextInt(world.board.height))
	      if (plant(freepos)) {
	        attempts=0
	      }
      }
    }
  }

  def plant(pos : Pos): Boolean= {
    if (world.board.moveTo(pos,this)) {
      if (near != null) {
        near.drain(50)
      }
      //world.flora += this
      true
    } else
      false
  }

  def die() {
    world.clock -= grow _
    world.newDay -= get_older _
    shouldExpand -= expand _
    shouldDie -= die _
  } 


  /** the plant expands, i.e. a new plant may germinate in its proximity */
  protected def expand() {
      new Plant(world,this,maxAge+10-plnt_rnd.nextInt(15))
      size-=300
  }

  def get_older() {
    age += 1
  }

  def canLive = (age <= maxAge) && (size <= 3000) && (size >= 1)

  def grow() {
    if (size >= 1) { 
      val range_limit=6
      val neighbours_all=world.board.neighbours(this,range_limit)
      val neighbours_plt = neighbours_all.filter({
        case p: Plant if p != this => true
        case _ => false
      }) map { case p: Plant => p }
      
      var neighbours_total:Int=0
      neighbours_plt.foreach(p => neighbours_total += (p.size / (1 + p.distance(this))).abs)
      if (neighbours_total>6000 && age>30) {
        killed()
      }
      val upper=80
      if (neighbours_total>upper) {
        maxAge-=5  
        neighbours_total=upper
      }
      size += 5+plnt_rnd.nextInt(1+(10/(1+neighbours_total)).toInt)
    }
  }
  
  
  override val symbol="^"
  
  override def toString =
    "P(" + (size.toFloat / 10) + ")#" + age + "@" + pos

}


class Gardener(val world : World)  {
  world.plantDied += prune _
  def prune(plant : Plant) {
    val board = plant.board
    if (board != null) {
      plant.board = null
    }
  }
}


