package universe

import plant._
import animal._
import handler._

class World(val width:Int=20,val height:Int=20)  extends AnimalDiedHandler 
  with PlantDiedHandler 
  with PlantPlantGerminatedHandler 
  with PlantPlantDisappearedHandler 
  with AddedElementHandler
  with WorldMonthHandler 
  with WorldWeekHandler 
  with WorldDayHandler 
  with WorldClockHandler {

  val board = new Board(width,height)
  val flora = new scala.collection.mutable.HashSet[Plant] 
  val animals = new scala.collection.mutable.HashSet[Animal] 

  //  evt plantDied = flora.any(p => p.disappeared)  
  //  evt plantGerminated = flora.any(p => p.germinated)  
  //
  //  evt animalDied = animals.any(p => p.disappeared)  
  //  evt animalBorn = animals.any(p => p.born)  

  // @JVH not used since the ones from the subclasses are used
  def handle_disappeared(be:BoardElement) = {
    // not used only needed because the class Board needs it
  }


  def handle_plant_plantDisappeared(pl : Plant) = {
    if (flora.contains(pl)) {
      send_plantDied(pl)
    }
  }

  def handle_animal_disappeared(an: Animal) = {
    if (animals.contains(an)) {
      send_animalDied(an)
    }
  }

  def handle_born(an : Animal) = {
    send_animalBorn(an)
  }

  def handle_plant_plantGerminated(pl : Plant) = {
    send_plantGerminated(pl)
  }

  // begin plantGerminated
  val plantGerminated_subscribers=new scala.collection.mutable.HashSet[PlantGerminatedHandler] 
  def subscribe_plantGerminated(subscriber:PlantGerminatedHandler)={
    plantGerminated_subscribers+=subscriber
  }
  def unsubscribe_plantGerminated(subscriber:PlantGerminatedHandler)={
    plantGerminated_subscribers-=subscriber
  }
  def send_plantGerminated(pl : Plant) ={
    plantGerminated_subscribers.foreach(_.handle_plantGerminated(pl))
  }
  // end plantGerminated

  // begin plantDied
  val plantDied_subscribers=new scala.collection.mutable.HashSet[PlantDiedHandler] 
  def subscribe_plantDied(subscriber:PlantDiedHandler)={
    plantDied_subscribers+=subscriber
  }
  def unsubscribe_plantDied(subscriber:PlantDiedHandler)={
    plantDied_subscribers-=subscriber
  }
  def send_plantDied(pl : Plant) ={
    plantDied_subscribers.foreach(_.handle_plantDied(pl))
  }
  // end plantDied

  // begin animalBorn
  val animalBorn_subscribers=new scala.collection.mutable.HashSet[AnimalBornHandler] 
  def subscribe_animalBorn(subscriber:AnimalBornHandler)={
    animalBorn_subscribers+=subscriber
  }
  def unsubscribe_animalBorn(subscriber:AnimalBornHandler)={
    animalBorn_subscribers-=subscriber
  }
  def send_animalBorn(an : Animal) ={
    animalBorn_subscribers.foreach(_.handle_animalBorn(an))
  }
  // end animalBorn

  // begin animalDied
  val animalDied_subscribers=new scala.collection.mutable.HashSet[AnimalDiedHandler] 
  def subscribe_animalDied(subscriber:AnimalDiedHandler)={
    animalDied_subscribers+=subscriber
  }
  def unsubscribe_animalDied(subscriber:AnimalDiedHandler)={
    animalDied_subscribers-=subscriber
  }
  def send_animalDied(an : Animal) ={
    animalDied_subscribers.foreach(_.handle_animalDied(an))
  }
  // end animalDied


  var hour : Int =0
  var day : Int = 0
  var week : Int =0

  // imperative evt clock[Unit]   @JVH

  // @JVH begin  event clock
  val world_clock_subscribers=new scala.collection.mutable.HashSet[WorldClockHandler]

  def subscribe_world_clock(subscriber:WorldClockHandler) = {
    world_clock_subscribers+=subscriber
  }

  def unsubscribe_world_clock(subscriber:WorldClockHandler) = {
    world_clock_subscribers-=subscriber
  }

  def send_world_clock() = {
    world_clock_subscribers.foreach(_.handle_world_clock)
  }
  // @JVH end


  // @JVH begin event newDay
  val world_day_subscribers=new scala.collection.mutable.HashSet[WorldDayHandler]

  def subscribe_world_day(subscriber:WorldDayHandler) = {
    world_day_subscribers+=subscriber
  }

  def unsubscribe_world_day(subscriber:WorldDayHandler) = {
    world_day_subscribers-=subscriber
  }

  def send_world_day() = {
    world_day_subscribers.foreach(_.handle_world_newDay)
  }
  // @JVH end


  // @JVH begin event newWeek
  val world_week_subscribers=new scala.collection.mutable.HashSet[WorldWeekHandler]

  def subscribe_world_week(subscriber:WorldWeekHandler) = {
    world_week_subscribers+=subscriber
  }

  def unsubscribe_world_week(subscriber:WorldWeekHandler) = {
    world_week_subscribers-=subscriber
  }

  def send_world_week() = {
    world_week_subscribers.foreach(_.handle_world_newWeek)
  }
  // @JVH end

  // @JVH begin event newWeek
  val world_month_subscribers=new scala.collection.mutable.HashSet[WorldMonthHandler]

  def subscribe_world_month(subscriber:WorldMonthHandler) = {
    world_month_subscribers+=subscriber
  }

  def unsubscribe_world_month(subscriber:WorldMonthHandler) = {
    world_month_subscribers-=subscriber
  }

  def send_world_month() = {
    world_month_subscribers.foreach(_.handle_world_newMonth)
  }
  // @JVH end


  board.subscribe_addedElement(this)

  this.subscribe_world_clock(this)
  def handle_world_clock() = {
    this.nextHour
    if (hour==0)
      send_world_day()
  }

  this.subscribe_world_day(this)
  def handle_world_newDay() = {
    this.nextDay
    if (day==0)
      send_world_week()
  }

  this.subscribe_world_week(this)
  def handle_world_newWeek() = {
    this.nextWeek
    if (week==0)
      send_world_month()
  }

  this.subscribe_world_month(this)
  def handle_world_newMonth() = {
    this.nextMonth
  }

//  evt newDay[Unit] = clock && (() => hour == 0)
//  evt newWeek[Unit] = newDay && (() => day == 0)
//  evt newMonth[Unit] = newWeek && (() => week == 0)

//  clock += nextHour _
//  newDay += nextDay _
//  newWeek += nextWeek _
//  newMonth += nextMonth _

  subscribe_animalDied(this)
  subscribe_plantDied(this)

  //  animalDied += onAnimalDied _
  //  plantDied += onPlantDied _

  def handle_addedElement(be : BoardElement) = {
    be match {
      case a:Animal =>  
	animals += a
      case p:Plant  => 
	flora += p
    }
  }


  def handle_animalDied(a: Animal) {
    animals -= a
  }


  def handle_plantDied(p: Plant) {
    flora -= p
  }

  def dump = {
    println("Dump "+board.surface.size + ". Week: " + week + ", Day: " + day + ", Hour: " + hour)
    for (y <- 0 to height) {
      var line=""
      var img=""
      for (x <- 0 to width) {
        val e=board(x,y)
        if (e!=null) {
          line+=" "+e.toString
          img+=e.symbol
        } else {
          img += " "
        }
      }
      //      println("|"+img+"|"+line)
      println("|"+img+"|")
    }
  }

  def nextHour() {
    hour = (hour + 1) % 24
  }

  def nextDay() {
    day = (day + 1) % 7
  }

  def nextWeek() {
    week = (week + 1) % 5
    new Plant(this, null, 150)
  }

  def nextMonth() {
  }

}

object Test {
  val wrld=new World(50,25)
  val gardener=new Gardener(wrld)
  
 def putH = {
    val board=wrld.board
    val center=new Pos(board.width/2,board.height/2)
    val inner=if (board.width>board.height) board.height else board.width
    for (i <-1 to 5) {
      val m_pos=center.random_around(3,inner/3)
      val f_pos=center.random_around(3,inner/3)
      
      board.moveTo(m_pos,
                   new Herbivore(wrld,5,70) with HerbivoreMale)
      board.moveTo(f_pos,
                   new Herbivore(wrld,3,80) with HerbivoreFemale)
      board.moveTo(center.random_around(inner/2, inner - 1),
        new Carnivore(wrld, 5, 100) with CarnivoreMale)
      board.moveTo(center.random_around(inner/2, inner -1),
        new Carnivore(wrld, 5, 100) with CarnivoreFemale)
      new Plant(wrld,null,80)
      new Plant(wrld,null,60)
      new Plant(wrld,null,40)
    }
  }


  def main(a : Array[String]) = {
    putH
    var count=1
    while (true) {
      wrld.send_world_clock()
      wrld.dump
      Thread.sleep(100)
      count+=1
      if (wrld.animals.size==0) {
	print ("count :"+count)
	exit
      }
    }
  }
}
