
package universe
package animal

// @JVH import scala.events._

import plant._
import universe.handler.AnimalDiedHandler
import universe.handler.PlantDiedHandler
import universe.handler.AnimalTargetReachedHandler
import universe.handler.AnimalPlantReachedHandler
import universe.handler.AnimalAnimalReachedHandler
import universe.handler.AnimalKilledHandler
import universe.handler.AnimalShouldDieHandler
import universe.handler.WorldClockHandler
import universe.handler.WorldDayHandler

object Animal {  
  def asAnimal(be:BoardElement): Option[Animal] = be match {
    case a: Animal => Some(a)
    case _ => None
  }
  var ac:Int =0
  def get_id:Int ={
    ac+=1
    return ac
  }
}

abstract class Animal(val world: World, var range:Int, val maxAge:Int) extends BoardElement 
	 with AnimalTargetReachedHandler 
	 with AnimalKilledHandler
	 with AnimalShouldDieHandler
	 with AnimalDiedHandler
	 with PlantDiedHandler
	 with WorldClockHandler 
	 with WorldDayHandler {
	 val id=Animal.get_id  
	   /** this event makes it possible to refine when an animal should move */
	   // @JVH ßevt movementTick[Unit] = world.clock
	   world.subscribe_world_clock(this)


	   subscribe_animal_born(world)
	   subscribe_animal_disappeared(world)

	   // @JVH begin animal_movementTick
	   // just skip the event here since it doesn't allow much modularity
	   //  multiple handlers are a bit hard to do in an understandable way
	   def handle_world_clock ={	     
	     received_movementTick_forward_1
	     movement
	   }
	   // @JVH end animal_movementTick



	   // @JVH begin animal_born
	   lazy val animal_born_subscribers=new scala.collection.mutable.HashSet[World]

	   def subscribe_animal_born(subscriber:World) = {
	     animal_born_subscribers+=subscriber
	   }

	   def unsubscribe_animal_born(subscriber:World) = {
	     animal_born_subscribers-=subscriber
	   }

	   def send_animal_born(an : Animal) = {
	     animal_born_subscribers.foreach(_.handle_born(an))
	   }
	   // @JVH end animal_born

	   // @JVH evt born[Animal] = afterExec(appear) map ((_: Any) => this) 
	   
	   // @JVH evt disappeared[Animal] = afterExec(die) map ((_: Any) => this)

	   // @JVH begin animal_disappeared  
	   lazy val animal_disappeared_subscribers=new scala.collection.mutable.HashSet[World]

	   def subscribe_animal_disappeared(subscriber:World) = {
	     animal_disappeared_subscribers+=subscriber
	   }

	   def unsubscribe_animal_disappeared(subscriber:World) = {
	     animal_disappeared_subscribers-=subscriber
	   }

	   def send_animal_disappeared(an : Animal) = {
	     animal_disappeared_subscribers.foreach(_.handle_animal_disappeared(an))
	   }
	   // @JVH end animal_disappeared



	   /** Internal event to unregister everything that need to be unregistered before notifying the death to outer */
	   // @JVH protected evt shouldDie[Unit] = afterExec(get_older) && (_ => !canLive) || killed


	   // @JVH begin  not so declaritive 
	   def handle_animalKilled() = {
	     send_animal_shouldDie
	   }

	   subscribe_animal_killed(this) // @JVH
	   // @JVH end


	   // @JVH begin
	   lazy val animal_shouldDie_subscribers=new scala.collection.mutable.HashSet[AnimalShouldDieHandler]

	   protected def subscribe_animal_shouldDie(subscriber:AnimalShouldDieHandler) = {
	     animal_shouldDie_subscribers+=subscriber
	   }

	   protected def unsubscribe_animal_shouldDie(subscriber:AnimalShouldDieHandler) = {
	     animal_shouldDie_subscribers-=subscriber
	   }

	   def send_animal_shouldDie() = {
	     animal_shouldDie_subscribers.foreach(_.handle_animalShouldDie())
	   }
	   // @JVH end


	   // @JVH imperative evt killed[Unit]
	   // @JVH begin
	   lazy val animal_killed_subscribers=new scala.collection.mutable.HashSet[AnimalKilledHandler]

	   def subscribe_animal_killed(subscriber:AnimalKilledHandler) = {
	     animal_killed_subscribers+=subscriber
	   }

	   def unsubscribe_animal_killed(subscriber:AnimalKilledHandler) = {
	     animal_killed_subscribers-=subscriber
	   }

	   def send_animal_killed() = {
	     animal_killed_subscribers.foreach(_.handle_animalKilled)
	   }
	   // @JVH end


	   
	   // @JVH evt targetReached[BoardElement] = (afterExec(movement) && ((_: Any) => reached)) map ((_: Any) => target()) 

	   // @JVH begin
	   lazy val animal_targetReached_subscribers=new scala.collection.mutable.HashSet[AnimalTargetReachedHandler]

	   def subscribe_animal_targetReached(subscriber:AnimalTargetReachedHandler) = {
	     animal_targetReached_subscribers+=subscriber
	   }

	   def unsubscribe_animal_targetReached(subscriber:AnimalTargetReachedHandler) = {
	     animal_targetReached_subscribers-=subscriber
	   }

	   def send_animal_targetReached(be : BoardElement) = {
	     animal_targetReached_subscribers.foreach(_.handle_animalTargetReached(be))
	   }
	   // @JVH end



	   // @JVH begin  comes from InteractWithAnimals
	   lazy val animal_animalReached_subscribers=new scala.collection.mutable.HashSet[AnimalAnimalReachedHandler]

	   def subscribe_animal_animalReached(subscriber:AnimalAnimalReachedHandler) = {
	     animal_animalReached_subscribers+=subscriber
	   }

	   def unsubscribe_animal_animalReached(subscriber:AnimalAnimalReachedHandler) = {
	     animal_animalReached_subscribers-=subscriber
	   }

	   def send_animal_animalReached(be : BoardElement) = {
	     Animal.asAnimal(be) match {
	       case None => throw new IllegalArgumentException("arg not Animal")
	       case Some(animal) => animal_animalReached_subscribers.foreach(_.handle_animalAnimalReached(animal))
	     }
	   }
	   // @JVH end


	   // @JVH begin  comes from InteractWithPlants
	   lazy val animal_plantReached_subscribers=new scala.collection.mutable.HashSet[AnimalPlantReachedHandler]

	   def subscribe_animal_plantReached(subscriber:AnimalPlantReachedHandler) = {
	     animal_plantReached_subscribers+=subscriber
	   }

	   def unsubscribe_animal_plantReached(subscriber:AnimalPlantReachedHandler) = {
	     animal_plantReached_subscribers-=subscriber
	   }

	   def send_animal_plantReached(be : BoardElement) = {
	     Plant.asPlant(be) match {
	       case None => throw new IllegalArgumentException("arg not Plant")
	       case Some(plnt) => animal_plantReached_subscribers.foreach(_.handle_animalPlantReached(plnt))
	     }
	   }
	   // @JVH end

	   // @JVH  translation of the newTargetNeeded conditions and reactions
	   def handle_animalTargetReached(be:BoardElement) = {
	     val animal=Animal.asAnimal(be)
	     val plant=Plant.asPlant(be)
	     if (animal!=None)
	       send_animal_animalReached(be)
	     if (plant!=None)
	       send_animal_plantReached(be)

	     nextTarget // @JVH part of newTargetNeeded
	   }

	   // forwarded because we received a movementTick message 
	   def received_movementTick_forward_1 = {
	     if (target==null) {
	       nextTarget
	     }
	   }
	   
	   def handle_animalDied(an:Animal) ={
	     if (an==target)
	       nextTarget
	   }
	   def handle_plantDied(pl:Plant) ={
	     if (pl==target)
	       nextTarget
	   }
	   // @JVH


	   // @JVH  protected evt newTargetNeeded[Unit] = targetReached || target.event(_.disappeared) || (movementTick && (() => target == null))

	   val isMale: Boolean

	   var age = 0
	   var energy = 15 
	   val maxRange = 1 + range / 12

	   // try to move to this position
	   //  @JVH protected lazy val target: Variable[BoardElement] = Variable(null)
	   protected var target: BoardElement = null


	   def appear() {

	     // @JVH  movementTick += movement _ 
	     // @JVH  world.newDay += get_older _
	     world.subscribe_world_day(this) // @JVH
	     
	     // @JVH newTargetNeeded += nextTarget _    no longer an event
	     // @JVH shouldDie += die _
	     subscribe_animal_shouldDie(this) // @JVH
	     subscribe_animal_targetReached(this) // @JVH for newTargetNeeded

	     world.subscribe_animalDied(this) // @JVH for newTargetNeeded
	     world.subscribe_plantDied(this) // @JVH for newTargetNeeded

	     send_animal_born(this) // @JVH
	   }

	   /** defines the next target and start moving to it */
	   def nextTarget() {
	     if (board!=null)  { // @JVH only when animal is on a board
	       target = find_target(maxRange) // @JVH
	     }
	   }

	   def handle_animalShouldDie() ={
	     die
	   }

	   def die() {
	     board = null
	     world.unsubscribe_world_clock(this) // @JVH
	     // @JVH movementTick -= movement _  
	     // @JVH world.newDay -= get_older _
	     world.unsubscribe_world_day(this) // @JVH
	     // @JVH  newTargetNeeded -= nextTarget _        no longer an event
	     // @JVH shouldDie -= die _
	     unsubscribe_animal_shouldDie(this) // @JVH
	     unsubscribe_animal_killed(this) // @JVH
	     unsubscribe_animal_targetReached(this) // @JVH for newTargetNeeded

	     send_animal_disappeared(this) // @JVH
	   } 
	   
	   def isFertile: Boolean =
	     age>10 && energy>40

	   // check if the animal can still live
	   def canLive = age < maxAge && energy > 0

	   // @JVH begin
	   def handle_world_newDay() {
	     get_older()
	   }
	   // @JVH end
	   
	   def get_older() {
	     age+=1
	     energy-=2 
	     after_get_older // @JVH
	   }

	   // @JVH begin  (after get_older)	
	   def after_get_older()={
	     if (!canLive) {
	       send_animal_shouldDie
	     }
	   }
	   // @JVH end
	   
	   // abstract method to find out where to go
	   def find_target(range :Int):BoardElement 


	   // try to move towards the target
	   // for now allow jumping over
	   def movement() {
	     
	     val targetpos = if (target==null) { // @JVH 
	       pos.random_around(1+maxRange,1+maxRange*2.toInt) 
	     } else { 
	       target.pos // @JVH 
	     }
	     
	     val delta=pos-targetpos
	     var movement_x=0
	     var movement_y=0
	     var movement:Pos=null
	     val steps=scala.math.sqrt(maxRange).toInt
	     
	     if (delta.x < 0) {
	       movement_x=1 }
	     if (delta.x > 0) {
	       movement_x=0-1}
	     if (delta.y < 0) {
	       movement_y=1}
	     if (delta.y > 0) {
	       movement_y=0-1}
	     if (movement_x*movement_x+movement_y*movement_y>1) {
	       movement=new Pos(movement_x*steps,movement_y*steps)
	     } else {
	       movement=new Pos(movement_x*maxRange,movement_y*maxRange)
	     }   
	     
	     val move_max=movement.distance
	     val move_needed=delta.distance
	     
	     if (move_max>move_needed) {
	       if (move_max>0) {
		 movement=movement*(move_needed.toDouble/move_max)
	       }
	     }
	     if (moveTo(pos+movement)) {
	       energy -= energyLossFactor * movement.distance.toInt
	     }
	     after_movement // @JVH
	   }

	   // @JVH begin check if we need to send the event after testing
	   def after_movement() ={
	     if (reached) {
	       send_animal_targetReached(target) 
	     }
	   }
	   // @JVH  end

	   protected def energyLossFactor = 1/2

	   def reached() =
	     target != null && distance(target) < 2 // @JVH

	   override def symbol="A"

	   override def toString = symbol+"(" + (energy.toFloat / 10) + ")#" + age + "@" + pos

	   
	   subscribe_animal_born(world) // @JVH
	 }
