package universe
package animal

import universe.handler.AnimalAnimalReachedHandler
import universe.handler.AnimalSameSpeciesReachedHandler
import universe.handler.AnimalOtherSpeciesReachedHandler

trait InteractWithAnimals extends Animal with AnimalAnimalReachedHandler  {

  

  /** global event triggered when another animal is reached */
//  evt animalReached[Animal] = (targetReached && (_.isInstanceOf[Animal])) map ((be: BoardElement) => Animal.asAnimal(be).get) // @JVH
  // the event above had to be moved to Animal.scala

  /** event triggered when an animal of the same species is reached */
  // @JVH evt sameSpeciesReached[Animal] = animalReached && sameSpecies _

  /** event triggered when an animal of another species is reached */
  // @JVH evt otherSpeciesReached[Animal] = animalReached && (a => !sameSpecies(a))

  // @JVH begin  
  val animal_sameSpeciesReached_subscribers=new scala.collection.mutable.HashSet[AnimalSameSpeciesReachedHandler]

  def subscribe_animal_SameSpeciesReached(subscriber:AnimalSameSpeciesReachedHandler) = {
    animal_sameSpeciesReached_subscribers+=subscriber
  }

  def unsubscribe_animal_SameSpeciesReached(subscriber:AnimalSameSpeciesReachedHandler) = {
    animal_sameSpeciesReached_subscribers-=subscriber
  }

  def send_animal_sameSpeciesReached(animal:Animal) = {
    animal_sameSpeciesReached_subscribers.foreach(
      _.handle_animalSameSpeciesReachedHandler(animal)
    )
  }
  // @JVH end


  // @JVH begin  
  val animal_otherSpeciesReached_subscribers=new scala.collection.mutable.HashSet[AnimalOtherSpeciesReachedHandler]

  def subscribe_animal_OtherSpeciesReached(subscriber:AnimalOtherSpeciesReachedHandler) = {
    animal_otherSpeciesReached_subscribers+=subscriber
  }

  def unsubscribe_animal_OtherSpeciesReached(subscriber:AnimalOtherSpeciesReachedHandler) = {
    animal_otherSpeciesReached_subscribers-=subscriber
  }

  def send_animal_otherSpeciesReached(animal:Animal) = {
    animal_otherSpeciesReached_subscribers.foreach(
      _.handle_animalOtherSpeciesReachedHandler(animal)
    )
  }
  // @JVH end



  // @JVH begin   less declarative but still understandable
  def handle_animalAnimalReached(animal : Animal) {
    if (sameSpecies(animal)) 
      send_animal_sameSpeciesReached(animal)
    else 
      send_animal_otherSpeciesReached(animal)
  }
  // @JVH end

  def sameSpecies(a: Animal): Boolean

}

// vim: set ts=4 sw=4 et:
