package universe
package gui

import scala.swing._
import javax.swing.ImageIcon

abstract class GuiBoard(world: World) extends Table {
  import Table._

  showGrid = true
  rowHeight = 40
  model = new BoardModel(world)
  
  override def apply(row: Int, col: Int): BoardElement =
    super.apply(row, col) match {
      case be: BoardElement => be
      case _ => null // SHOULD NOT HAPPEN
    }

  override def update(row: Int, col: Int, v: Any): Unit =
    v match {
      case be: BoardElement => super.update(row, col, be)
      case _ => throw new IllegalArgumentException("Only BoardElement can be added to the board")
    }

  override def rendererComponent(isSelected: Boolean, hasFocus: Boolean, row: Int, col: Int): Component = {
    renderElement(this(row, col))
  }

  object emptyImage extends ImageIcon(getClass.getResource("/img/empty.png"))

  def renderElement(be: BoardElement): Component =
    if(be == null)
      new Label("", emptyImage, Alignment.Center)
    else
      new Label(be.symbol, emptyImage, Alignment.Center)

}

// vim: set ts=4 sw=4 et:
