package universe
package plant

import universe.handler.PlantDiedHandler
import universe.handler.PlantPlantGerminatedHandler
import universe.handler.PlantPlantDisappearedHandler
import universe.handler.PlantKilledHandler
import universe.handler.PlantShouldDieHandler
import universe.handler.PlantShouldExpandHandler
import universe.handler.WorldClockHandler
import universe.handler.WorldDayHandler
import universe.handler.WorldWeekHandler

object Plant {

  def asPlant(b:BoardElement): Option[Plant] = b match {
    case p: Plant => Some(p)
    case _ => None
  }
  
}

class Plant(val world: World,val near:Plant =null,var maxAge:Int=160) 
extends BoardElement
with PlantKilledHandler 
with PlantShouldDieHandler 
with PlantShouldExpandHandler 
with WorldClockHandler 
with WorldDayHandler {
  val plnt_rnd = new scala.util.Random
  


  // @JVH  evt germinated[Plant] = afterExec(plant) && ((_: Pos, success: Boolean) => success) map ((_: Any) => this)

  // @JVH begin germinated
  lazy val plant_germinated_subscribers=new scala.collection.mutable.HashSet[PlantPlantGerminatedHandler]
  
  def subscribe_plant_plantGerminated(subscriber:PlantPlantGerminatedHandler) = {
    plant_germinated_subscribers+=subscriber
  }
  
  def unsubscribe_plant_plantGerminated(subscriber:PlantPlantGerminatedHandler) = {
    plant_germinated_subscribers-=subscriber
  }
  
  def send_plant_plantGerminated(plnt : Plant) = {
    plant_germinated_subscribers.foreach(_.handle_plant_plantGerminated(plnt))
  }
  // @JVH end germinated


  // @JVH begin disappeared
  lazy val plant_disappeared_subscribers=new scala.collection.mutable.HashSet[PlantPlantDisappearedHandler]
  
  def subscribe_plant_plantDisappeared(subscriber:PlantPlantDisappearedHandler) = {
    plant_disappeared_subscribers+=subscriber
  }
  
  def unsubscribe_plant_plantDisappeared(subscriber:PlantPlantDisappearedHandler) = {
    plant_disappeared_subscribers-=subscriber
  }
  
  def send_plant_plantDisappeared(plnt : Plant) = {
    plant_disappeared_subscribers.foreach(_.handle_plant_plantDisappeared(plnt))
  }
  // @JVH end disappered

  
  // @JVH evt disappeared[Plant] = afterExec(die) map ((_: (Unit, Unit)) => this)

  /** Internal event to unregister everything that need to be unregistered before notifying the death to outer */
  // @JVH protected evt shouldDie[Unit] = (afterExec(get_older) || afterExec(grow)) && (_ => !canLive) || killed

  // @JVH imperative evt killed[Unit]
  
  // @JVH begin
  lazy val plant_killed_subscribers=new scala.collection.mutable.HashSet[PlantKilledHandler]
  
  def subscribe_plant_killed(subscriber:PlantKilledHandler) = {
    plant_killed_subscribers+=subscriber
  }
  
  def unsubscribe_plant_killed(subscriber:PlantKilledHandler) = {
    plant_killed_subscribers-=subscriber
  }
  
  def send_plant_killed() = {
    plant_killed_subscribers.foreach(_.handle_plantKilled)
  }
  // @JVH end


  // @JVH begin
  lazy val plant_shouldDie_subscribers=new scala.collection.mutable.HashSet[PlantShouldDieHandler]
  
  def subscribe_plant_shouldDie(subscriber:PlantShouldDieHandler) = {
    plant_shouldDie_subscribers+=subscriber
  }
  
  def unsubscribe_plant_shouldDie(subscriber:PlantShouldDieHandler) = {
    plant_shouldDie_subscribers-=subscriber
  }
  
  def send_plant_shouldDie() = {
    plant_shouldDie_subscribers.foreach(_.handle_plantShouldDie)
  }
  // @JVH end

  // @JVH begin
  lazy val plant_shouldExpand_subscribers=new scala.collection.mutable.HashSet[PlantShouldExpandHandler]
  
  def subscribe_plant_shouldExpand(subscriber:PlantShouldExpandHandler) = {
    plant_shouldExpand_subscribers+=subscriber
  }
  
  def unsubscribe_plant_shouldExpand(subscriber:PlantShouldExpandHandler) = {
    plant_shouldExpand_subscribers-=subscriber
  }
  
  def send_plant_shouldExpand() = {
    plant_shouldExpand_subscribers.foreach(_.handle_plantShouldExpand)
  }
  // @JVH end




  // @JVH protected evt shouldExpand[Unit] = afterExec(get_older) && ((_: Any) => (age > 10 && size > 60 && plnt_rnd.nextInt(50) >= 30))

  var age = 0
  var size = 1 
  override def move(r: Int) = false

  // @JVH begin
  def handle_world_clock={
    grow 
  }
  // @JVH end

  def appear() {
    subscribe_plant_killed(this) // @JVH for shouldDie
    // @JVH   world.clock += grow _
    world.subscribe_world_clock(this) // @JVH
    // @JVH world.newDay += get_older _
    world.subscribe_world_day(this)
    // @JVH shouldExpand += expand _
    subscribe_plant_shouldExpand(this) // @JVH
  }

  def drain(amount : Int) = {
    if (amount>size) {
      val oldSize=size
      size=0
      canLive
      oldSize
    } else {
      size -= amount
      amount
    }
  }

  def init {
    subscribe_plant_shouldDie(this) // @JVH
 
    this.subscribe_plant_plantGerminated(world) //  @JVH  plant.any workaround
    this.subscribe_plant_plantDisappeared(world) //  @JVH  plant.any workaround

    var attempts = 10
    while (attempts > 0) {
      attempts -= 1
      if (near != null) {
        val freepos = near.find_free_around(4 + size / 5, 2 + size / 8)
        if (freepos == null) {
          near.age += 50
          attempts = 0
        } else if (plant(freepos)) {
          attempts = 0
        } 
      } 
      if (near == null) {
	val freepos = new Pos(plnt_rnd.nextInt(world.board.width),
			      plnt_rnd.nextInt(world.board.height))
	if (plant(freepos)) {
	  attempts=0
	}
      }
    }
  }

  
  // @JVH begin
  def plant(pos : Pos): Boolean= {
    val res=plantImpl(pos) // get result for afterexec
    after_plant(res)
    return res
  }

  def plantImpl(pos : Pos): Boolean= {
    if (world.board.moveTo(pos,this)) {
      if (near != null) {
        near.drain(50)
      }
      //world.flora += this
      true
    } else
      false
  }

  def after_plant(res : Boolean)={
    if (res) {
      send_plant_plantGerminated(this)
    }
  }
  // @JVH end

  def handle_plantShouldDie()={
    die()
  }

  def die() = {
    // @JVH    world.clock -= grow _
    world.unsubscribe_world_clock(this) // @JVH
    // @JVH    world.newDay -= get_older _
    world.unsubscribe_world_day(this) // @JVH
    unsubscribe_plant_killed(this) // @JVH for shouldDie
    unsubscribe_plant_shouldDie(this) // @JVH
    unsubscribe_plant_shouldExpand(this) // @JVH
    
    // @JVH    shouldExpand += expand _    is this correct??
    after_die()
  } 

  def after_die() = {
    send_plant_plantDisappeared(this)
  }

  def handle_plantShouldExpand()={
    expand()
  }

  /** the plant expands, i.e. a new plant may germinate in its proximity */
  protected def expand() {
    new Plant(world,this,maxAge+10-plnt_rnd.nextInt(15))
    size-=45
  }


  // @JVH begin 
  def handle_world_newDay() = {
    this.get_older
  }
  // @JVH end

  def get_older() {
    age += 1
    after_get_older
  }

  // @JVH begin 
  def after_get_older={
    //  part of declarative shouldDie event
    if (!canLive) { 
      send_plant_shouldDie()
    }

    //  part of declarative shouldExpand event
    if (age >10 && size>60 && plnt_rnd.nextInt(50) >= 20) {
      send_plant_shouldExpand()
    }
  }
  // @JVH end  

  def canLive = (age <= maxAge) && (size <= 3000) && (size >= 1)

  def grow() {
    if (size >= 1) { 
      val range_limit=6
      val neighbours_all=world.board.neighbours(this,range_limit)
      val neighbours_plt = neighbours_all.filter({
        case p: Plant if p != this => true
        case _ => false
      }) map { case p: Plant => p }
      
      var neighbours_total:Int=0
      neighbours_plt.foreach(p => neighbours_total += (p.size / (1 + p.distance(this))).abs)
      if (neighbours_total>6000 && age>30) {
        send_plant_killed()
      }
      val upper=80
      if (neighbours_total>upper) {
        maxAge-=5  
        neighbours_total=upper
      }
      size += 5+plnt_rnd.nextInt(1+(10/(1+neighbours_total)).toInt)
    }
    after_grow() // @JVH
  }

  // @JVH begin part of declarative shouldDie event   
  def after_grow() = {
    if (!canLive) {
      send_plant_shouldDie()
    }
  }
  // @JVH end  

  // @JVH begin part of declarative shouldDie event   
  def handle_plantKilled={
    send_plant_shouldDie
  }
  // @JVH end
  
  override val symbol="^"
  
  override def toString =
    "P(" + (size.toFloat / 10) + ")#" + age + "@" + pos


  init

}


class Gardener(val world : World)  extends PlantDiedHandler {
  //  world.plantDied += prune _
  world.subscribe_plantDied(this)
  def handle_plantDied(plant : Plant) = {
    prune(plant)
  }
  def prune(plant : Plant) = {
    val board = plant.board
    if (board != null) {
      plant.board = null
    }
  }
}


